import { ANY, ARRAY } from '../constants.js';

// import udomdiff from 'udomdiff';
import udomdiff from '../domdiff.js';

import Fragment from '../classes/fragment.js';
const { diff } = Fragment;

import { empty, valueOf } from '../utils.js';

const remove = node => node.remove();

/** @typedef {ANY | ARRAY | import("../constants.js").OBJECT} HINT */

/**
 * @param {Text} node
 * @param {string?} prev
 * @returns {(curr:string?) => void}
 */
const any = (node, prev) => curr => {
  if (prev !== curr) {
    prev = curr;
    node.data = curr ?? '';
  }
};

/**
 * @param {Comment} node
 * @param {Node[]} prev
 * @returns {(curr:Node[]) => void}
 */
const array = (node, prev) => curr => {
  prev = udomdiff(prev, curr, diff, node);
};

/**
 * @param {Node} prev
 * @returns {(curr:Node) => void}
 */
const object = prev => curr => {
  if (prev !== curr) {
    prev.replaceWith(valueOf(curr));
    prev = curr;
  }
};

/**
 * @param {Element} node
 * @param {HINT} hint
 * @returns
 */
const multi = (node, hint) => (
  hint === ARRAY ?
    array(node, empty) :
    (hint === ANY ?
      any(node, '') :
      object(node)
    )
);

/**
 * @param {Element} node
 * @param {HINT} hint
 * @returns {(value: any) => void}
 */
const oneOff = (node, hint) => value => {
  multi(node, hint)(value);
  if (hint === ARRAY) node.data = `[${value.length}]`;
};

/**
 * @param {Node} node
 * @param {HINT} hint
 * @param {boolean} once
 */
export default (node, hint, once) => (once ? oneOff : multi)(node, hint);
